% ICBV 2009 - Final Project
% authors: Ramy Ben Aroya  (049822927) & Ron Cohen (040169245)

function [output] = DetectGrapes (I)
% This function gets an RGB image and returns a binary matrix output.
% 1 - indicates that there is a grape on that pixel, and 0 otherwise.
    DB = getDBMask(I,100);
    I = double(rgb2gray(I));
    cannyEdges = getEdges(I,'canny',50,25);
    filtered = grapesFilter(I,DB);
    centers = getCentersByFilter(cannyEdges,filtered);
    output = drawCircles(centers); 
end

function[DB] = getDBMask(I,wndSize)
% This function gets an Image - I and a window size - wndSize and return
% the Dark/Bright map of the image. Dark - 0 Bright - 1
    darkEnvR = 67.2538;
    darkEnvG = 63.9540;
    darkEnvB = 58.8888;
    avgDark = (darkEnvR + darkEnvG + darkEnvB)/3;
    lightEnvR = 160.9342;
    lightEnvG = 152.0336;
    lightEnvB = 142.0486;
    avgLight = (lightEnvR + lightEnvG + lightEnvB)/3;
    
    R = I(:,:,1);
    G = I(:,:,2);
    B = I(:,:,3);
    
    [n,m] = size(R);
    DB = zeros(n,m);
    for i = 1:wndSize:n
        for j = 1:wndSize:m
            wndR = R(i:min(i+wndSize-1,n),j:min(j+wndSize-1,m));
            wndG = G(i:min(i+wndSize-1,n),j:min(j+wndSize-1,m));
            wndB = B(i:min(i+wndSize-1,n),j:min(j+wndSize-1,m));
            avg = ( mean(wndR(:)) + mean(wndG(:)) + mean(wndB(:)) )/3;
            disDark = abs(avg-avgDark);
            disLight = abs(avg-avgLight);
            if(disLight <= disDark)
                DB(i:min(i+wndSize,n),j:min(j+wndSize,m)) = 1;
            end
        end
    end
end

function [ bw ] = getEdges( I,method,wndSize1,wndSize2 )
% This function gets an image - I, a string of an edge detection method
% and two window sizes.
% The function returns an edge map. The edges are calculated in each window
% seperatly by the first window size, and then they are combined to a whole
% image.
% After that the functions scans windows by the second window size and
% cleans windows with high density of edge elements.

    densityRatioThreshold = 0.4;
    
    [n,m] = size(I);
    indI = 1:wndSize1:n;
    indJ = 1:wndSize1:m;
    bw = zeros(n,m);
    for i = indI
        for j = indJ
            wnd = I(i:min(i+wndSize1,n),j:min(j+wndSize1,m));
            wndBW = edge(wnd,method);
            bw(i:min(i+wndSize1,n),j:min(j+wndSize1,m)) = wndBW;
        end
    end
    % High density cleanup
    indI = 1:wndSize2:n;
    indJ = 1:wndSize2:m;    
    for i = indI
        for j = indJ
            wnd = bw(i:min(i+wndSize2,n),j:min(j+wndSize2,m));
            [wndN,wndM] = size(wnd);
            wndPos =wnd(wnd==1);
            posNum = length(wndPos(:));
             if(posNum/wndN/wndM > densityRatioThreshold)
                bw(i:min(i+wndSize2,n),j:min(j+wndSize2,m)) = 0;
             end
        end
    end    
end

 function [ filtered] = grapesFilter (I,DB) 
% This function returns a filter map of the image based on
% average-gradient-magnitude 
% I is a grayscale image
% DB is a map of dark and bright regions on the original RGB image of I

    avgI= mean(mean(I));
    ratioThreshold = 10;
    [fx,fy] = gradient(I);
    output = sqrt(fx.^2+fy.^2);
    grapesAvgDark=5;
    grapesAvgBright=8.5;
    eps = 1.5;
    [n,m]=size(output);
    wndSize=50;
    filtered = zeros(n,m);
    for i= 1 :wndSize:n
        for j=1 :wndSize :m
            window=output(i:min(i+wndSize,n),j:min(j+wndSize,m));
            avg=mean(window(:));
            %Dark case
            if (avg/avgI < ratioThreshold)
                if (DB(i,j) == 0)
                    if (avg  > grapesAvgDark -eps & avg < grapesAvgDark+eps) 
                        filtered(i:min(i+wndSize,n),j:min(j+wndSize,m))=1;
                    end
                else
                    %Bright case
                    if (avg  > grapesAvgBright -eps & avg < grapesAvgBright+eps) 
                        output(i:min(i+wndSize,n),j:min(j+wndSize,m))=1;
                    end
                end
            end
        end
    end
 end
 function [centers] = getCentersByFilter(I,f)
 % This function gets a grayscale image - I and a filtered map of potential
 % grapes area - f.
 % The function returns the matrix 'centers' that every pixel in it has
 % a value, which indicates the size of the circle's radius that the pixel
 % is the center of it.
    
    [n,m] = size(I);
    minRad = 7;
    maxRad = 15;
    stepRad = 1;
    minPartOfCircle = 0.25;
    radius = minRad:stepRad:maxRad;
    numRad = length(radius);
  
    centers = zeros(n,m);
    for i = 1:n
        for j = 1:m
            if (I(i,j) == 0& f(i,j)>0)
                sums = zeros(1,numRad);
                for r = 1:numRad
                    sums(r) =  getSumOnRadius(I,i,j,radius(r));
                end
                avgs = sums./radius/6;
                [maxAvg,maxIndex] = max(avgs);
                if (maxAvg >  minPartOfCircle)
                    centers(i,j) = radius(maxIndex);
                end
            end
        end
    end
 end
 
 function[s] = getSumOnRadius(I,i,j,r)
 % This function gets an image - I, a pixel place - (i,j) and a radius r.
 % The function returns the sum of the positive pixels, that has a distance r
 % from I(i,j)
    [n,m] = size(I);
    periPixNum = 6*r;
    angles = 0:2*pi/periPixNum:2*pi*(periPixNum-1)/periPixNum;
    iPixels = i+ ceil(r*sin(angles));
    jPixels = j+ ceil(r*cos(angles));
    iPixels(iPixels < 1) = 1;
    iPixels(iPixels > n) = n;
    jPixels(jPixels < 1) = 1;
    jPixels(jPixels > m) = m;
    s = 0;

    for k = 1:periPixNum
         s = s + I(iPixels(k),jPixels(k));
    end
end

function[out] = drawCircles(centers)
% This functions gets a matrix that each pixel in it is a center of a
% circle. The pixel value is the radius of circle.
% The function returns the matrix out, which is after the circle has been
% "colored"
    [n,m] = size(centers);
    out = zeros(n,m);
    out(centers > 0) = 1;
    for i =1:n
        for j = 1:m                
            if (centers(i,j) > 0)
                maxRad = centers(i,j);
                for r = 1:maxRad
                    periPixNum = ceil(6.2*r);
                    angles = 0:2*pi/periPixNum:2*pi*(periPixNum-1)/periPixNum;
                    iPixels = i+ ceil(r*sin(angles));
                    jPixels = j+ ceil(r*cos(angles));
                    iPixels(iPixels < 1) = 1;
                    iPixels(iPixels > n) = n;
                    jPixels(jPixels < 1) = 1;
                    jPixels(jPixels > m) = m;
                    
                    iPixels2 = i+ floor(r*sin(angles));
                    jPixels2 = j+ floor(r*cos(angles));
                    iPixels2(iPixels2 < 1) = 1;
                    iPixels2(iPixels2 > n) = n;
                    jPixels2(jPixels2 < 1) = 1;
                    jPixels2(jPixels2 > m) = m;


                    for k = 1:periPixNum
                         out(iPixels(k),jPixels(k)) = 1;
                         out(iPixels2(k),jPixels2(k)) = 1;
                    end
                end
            end
        end
    end
end